from enum import Enum
from typing import Optional

from pydantic import BaseModel, Field

from .tool import CrewAITool


class CrewAINodeType(str, Enum):
    AGENT = "Agent"
    TOOL = "Tool"
    CUSTOM_TOOL = "CustomTool"
    BASIC = "Basic"


class CrewAINode(BaseModel):
    name: str = Field(..., description="Name of the node")
    type: CrewAINodeType = Field(..., description="Type of the node")
    description: Optional[str] = Field(
        default=None,
        description="Short description of the node, used mainly for tool descriptions",
    )

    class Config:
        frozen = True  # Makes it hashable and immutable


class CrewAIEdge(BaseModel):
    start_node: str = Field(..., description="Name of the starting node")
    end_node: str = Field(..., description="Name of the end node")
    condition: Optional[str] = Field(
        default=None, description="Optional condition for the edge"
    )

    class Config:
        frozen = True  # Makes it hashable and immutable


class CrewAIGraph(BaseModel):
    name: str = Field(..., description="Name of the graph")
    nodes: set[CrewAINode] = Field(
        default_factory=set, description="List of nodes in the graph"
    )
    edges: set[CrewAIEdge] = Field(
        default_factory=set, description="List of edges in the graph"
    )

    def create_agents(self, agents: set[str]):
        for agent in agents:
            self.nodes.add(CrewAINode(name=agent, type=CrewAINodeType.AGENT))

    def create_tools(self, tools: list[CrewAITool]):
        for tool in tools:
            tool_name = tool.name
            tool_type = (
                CrewAINodeType.CUSTOM_TOOL if tool.custom else CrewAINodeType.TOOL
            )
            description = tool.description
            self.nodes.add(
                CrewAINode(name=tool_name, type=tool_type, description=description)
            )

    def connect_agent_to_tools(self, agent: str, tools: list[CrewAITool]):
        if not self.is_agent_in_graph(agent):
            return

        for tool in tools:
            tool_name = tool.name
            self.edges.add(
                CrewAIEdge(start_node=agent, end_node=tool_name, condition="tool_call")
            )
            self.edges.add(CrewAIEdge(start_node=tool_name, end_node=agent))

    def connect_agents(self, agent_connections: dict[str, list[str]]):
        for src_agent, dest_agents in agent_connections.items():
            if not self.is_agent_in_graph(src_agent):
                print(
                    f"Source agent node {src_agent} is missing in graph, cannot create an edge with it..."
                )
                continue

            condition = "hierarchical_delegation" if len(dest_agents) > 1 else None
            for dest_agent in dest_agents:
                if not self.is_agent_in_graph(src_agent):
                    print(
                        f"Destination agent node {dest_agent} is missing in graph, cannot create an edge with it..."
                    )
                    continue
                self.edges.add(
                    CrewAIEdge(
                        start_node=src_agent, end_node=dest_agent, condition=condition
                    )
                )

    def add_start_end_nodes(self, start_agents: list[str], end_agents: list[str]):
        self.nodes.add(CrewAINode(name="START", type=CrewAINodeType.BASIC))
        self.nodes.add(CrewAINode(name="END", type=CrewAINodeType.BASIC))
        for start_agent in start_agents:
            condition = "start" if len(start_agents) > 1 else None
            self.edges.add(
                CrewAIEdge(
                    start_node="START", end_node=start_agent, condition=condition
                )
            )
        for end_agent in end_agents:
            condition = "end" if len(start_agents) > 1 else None
            self.edges.add(
                CrewAIEdge(start_node=end_agent, end_node="END", condition=condition)
            )

    def is_agent_in_graph(self, agent: str):
        return CrewAINode(name=agent, type=CrewAINodeType.AGENT) in self.nodes



from enum import Enum
from typing import Dict, List, Optional

from pydantic import BaseModel, ConfigDict, Field


class NodeType(Enum):
    AGENT = "agent"
    BASIC = "basic"
    TOOL = "tool"
    CUSTOM_TOOL = "custom_tool"
    DEFAULT = "default"

    def __str__(self) -> str:
        return self.value


class ToolType(Enum):
    WEB_SEARCH = "web_search"
    LLM = "llm"
    CODE_INTERPRETER = "code_interpreter"
    DOCUMENT_LOADER = "document_loader"
    DEFAULT = "default"

    def __str__(self) -> str:
        return self.value


class VulnerabilityDefinition(BaseModel):
    name: str
    description: str
    security_framework_mapping: Dict[str, str]
    remediation: str


class NodeDefinition(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    node_type: NodeType = Field(..., alias="type")
    name: str
    description: Optional[str] = None
    label: Optional[str] = None
    category: Optional[ToolType] = None
    vulnerabilities: List[VulnerabilityDefinition] = Field(default_factory=list)


class EdgeDefinition(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    start: str
    end: str
    condition: Optional[str] = None


class GraphDefinition(BaseModel):
    name: str
    nodes: List[NodeDefinition]
    edges: List[EdgeDefinition]
    tools: List[NodeDefinition] = Field(default_factory=list)



import json
from importlib import resources
from typing import List, Literal

from pydantic import BaseModel

from ..graph import GraphDefinition, NodeType, VulnerabilityDefinition


class VulnerabilityMapDefinition(BaseModel):
    type: NodeType
    match: Literal["name", "category"]
    value: str
    vulnerabilities: List[VulnerabilityDefinition]


def load_vulnerabilities() -> List[VulnerabilityMapDefinition]:
    input_file = resources.files(__package__) / "vulnerabilities.json"
    with input_file.open("r", encoding="utf-8") as f:
        data = json.load(f)

    return [VulnerabilityMapDefinition.model_validate(v) for v in data]


def map_vulnerabilities(graph: GraphDefinition):
    """Map vulnerabilities to nodes and tools in the graph"""

    vulnerabilities = load_vulnerabilities()

    # Map vulnerabilities to nodes
    for node in graph.nodes:
        for vulnerability in vulnerabilities:
            if vulnerability.type == node.node_type:
                if vulnerability.match == "name" and vulnerability.value == node.name:
                    node.vulnerabilities.extend(vulnerability.vulnerabilities)
                elif (
                    vulnerability.match == "category"
                    and node.category
                    and vulnerability.value == node.category.value
                ):
                    node.vulnerabilities.extend(vulnerability.vulnerabilities)

    # Map vulnerabilities to tools
    for tool in graph.tools:
        for vulnerability in vulnerabilities:
            if vulnerability.type == tool.node_type:
                if vulnerability.match == "name" and vulnerability.value == tool.name:
                    tool.vulnerabilities.extend(vulnerability.vulnerabilities)
                elif (
                    vulnerability.match == "category"
                    and tool.category
                    and vulnerability.value == tool.category.value
                ):
                    tool.vulnerabilities.extend(vulnerability.vulnerabilities)



import subprocess
from typing import List

import pydot


class Graph(pydot.Dot):
    def __init__(self, nodes: List[pydot.Node], edges: List[pydot.Edge]):
        super().__init__(
            graph_type="digraph",
            fontname="Inter",
            layout="neato",
            overlap="prism",
            mode="sgd",
            outputorder="nodesfirst",
            sep="+50",
            labelloc="b",
            normalize=True,
        )
        self.set_node_defaults(fontsize=20, forcelabels=True, fontname="Inter")
        self.set_edge_defaults(fontsize=20, forcelabels=True, fontname="Inter")

        for node in nodes:
            self.add_node(node)

        for edge in edges:
            self.add_edge(edge)

    def generate(self) -> str:
        return subprocess.check_output(
            ["dot", "-Tsvg:cairo"],
            input=str(self).encode(),
        ).decode("utf-8")



import os
from typing import Generator, Optional


def walk_files(
    root_dir: str, extensions: set[str], blacklist: Optional[list[str]] = None
) -> Generator[str, None, None]:
    if blacklist is None:
        blacklist = [".venv"]
    for root, dirs, files in os.walk(root_dir):
        # Remove blacklisted directories from the search
        dirs[:] = [d for d in dirs if d not in blacklist]
        for file in files:
            _, file_ext = os.path.splitext(file)
            # Only yield files with certain extensions
            if file_ext in extensions:
                yield os.path.join(root, file)


def walk_python_files(
    root_dir: str, blacklist: Optional[list[str]] = None
) -> Generator[str, None, None]:
    """
    Walk through all Python files in a directory, optionally skipping blacklisted folders.
    :param root_dir: The base directory to start the search.
    :param blacklist: A list of folder names to skip. Defaults to ['.venv'].
    :return: A generator yielding paths to Python files.
    """
    if blacklist is None:
        blacklist = [".venv"]
    allowed_extensions = [".py"]
    yield from walk_files(
        root_dir=root_dir, extensions=set(allowed_extensions), blacklist=blacklist
    )


def walk_python_files_and_notebooks(
    root_dir: str, blacklist: Optional[list[str]] = None
) -> Generator[str, None, None]:
    """
    Walk through all Python files and Jupyter notebooks in a directory, optionally skipping blacklisted folders.
    :param root_dir: The base directory to start the search.
    :param blacklist: A list of folder names to skip. Defaults to ['.venv'].
    :return: A generator yielding paths to Python files and Jupyter notebooks.
    """
    if blacklist is None:
        blacklist = [".venv"]
    allowed_extensions = [".py", ".ipynb"]
    yield from walk_files(
        root_dir=root_dir, extensions=set(allowed_extensions), blacklist=blacklist
    )


def walk_yaml_files(
    root_dir: str, blacklist: Optional[list[str]] = None
) -> Generator[str, None, None]:
    """
    Walk through all YAML files in a directory, optionally skipping blacklisted folders.
    :param root_dir: The base directory to start the search.
    :param blacklist: A list of folder names to skip. Defaults to ['.venv'].
    :return: A generator yielding paths to YAML files.
    """
    if blacklist is None:
        blacklist = [".venv"]
    allowed_extensions = [".yaml", ".yml"]
    yield from walk_files(
        root_dir=root_dir, extensions=set(allowed_extensions), blacklist=blacklist
    )



import ast
from typing import Optional, Sequence, Union, cast


def find_return_of_function_call(
    node: Union[ast.AST, Sequence[ast.AST]], function_name: str
) -> Optional[ast.Call]:
    """
    Recursively search for and return the function call node contained in an expression of the form 'return {function_name}()'.
    Args:
        node: An AST node or sequence of AST nodes to search within
        function_name: The name of the function to search for
    Returns:
        The function call node if found, otherwise None
    """
    # Handle list of statements
    if isinstance(node, (list, tuple)):
        for item in node:
            result = find_return_of_function_call(item, function_name)
            if result:
                return result
        return None

    # Check if this is a return statement
    if isinstance(node, ast.Return):
        if node.value and is_function_call(node.value, function_name):
            return cast(ast.Call, node.value)  # Return the function call node
        return None

    # For compound statements, check their bodies
    elif isinstance(node, ast.If):
        # Check the if body - body is a list of statements
        result = find_return_of_function_call(node.body, function_name)
        if result:
            return result
        # Check the else body if it exists
        if node.orelse:
            return find_return_of_function_call(node.orelse, function_name)
        return None
    elif isinstance(node, ast.IfExp):
        # For expression-based if statements
        if_result = find_return_of_function_call(node.body, function_name)
        if if_result:
            return if_result
        return find_return_of_function_call(node.orelse, function_name)
    elif (
        isinstance(node, ast.For)
        or isinstance(node, ast.While)
        or isinstance(node, ast.With)
    ):
        return find_return_of_function_call(node.body, function_name)
    elif isinstance(node, ast.Try):
        # Check try body
        result = find_return_of_function_call(node.body, function_name)
        if result:
            return result
        # Check except handlers
        if node.handlers:
            for handler in node.handlers:
                result = find_return_of_function_call(handler.body, function_name)
                if result:
                    return result
        # Check finally block
        if node.finalbody:
            return find_return_of_function_call(node.finalbody, function_name)
        return None

    return None


def is_function_call(node: ast.AST, function_name: str) -> bool:
    """
    Check if a node is a call to a function with the given name.

    Args:
        node: An AST node to check
        function_name: The name of the function to check for

    Returns:
        True if the node is a call to the function, otherwise False
    """
    if isinstance(node, ast.Call):
        if isinstance(node.func, ast.Name) and node.func.id == function_name:
            return True
        elif isinstance(node.func, ast.Attribute) and node.func.attr == function_name:
            return True
    return False



from pathlib import Path

from ...analysis.analyze import Analyzer
from ...graph import EdgeDefinition as Edge
from ...graph import GraphDefinition, NodeType, ToolType
from ...graph import NodeDefinition as Node
from .custom_tools import get_all_custom_tools_from_directory
from .graph import parse_all_graph_instances_in_directory
from .predefined_tools import get_all_predefined_tools_from_directory
from .utils import build_global_registry

GRAPH_CLASS = "langgraph.graph.StateGraph"
COMMAND_CLASS = "langgraph.types.Command"
ADD_CONDITIONAL_EDGES_METHOD_NAME = "add_conditional_edges"
ADD_NODE_METHOD_NAME = "add_node"


class LangGraphAnalyzer(Analyzer):
    def __init__(self):
        super().__init__()

    def analyze(self, root_directory: str) -> GraphDefinition:
        global_functions, global_variables = build_global_registry(root_directory)

        predefined_tools = get_all_predefined_tools_from_directory(root_directory)

        custom_tools = get_all_custom_tools_from_directory(root_directory)

        all_graphs = parse_all_graph_instances_in_directory(
            root_directory,
            GRAPH_CLASS,
            COMMAND_CLASS,
            ADD_CONDITIONAL_EDGES_METHOD_NAME,
            ADD_NODE_METHOD_NAME,
            global_functions,
            global_variables,
        )
        nodes = []
        edges = []

        for graph in all_graphs:
            for i, node in enumerate(graph["graph_info"]["nodes"]):
                nodes.append(
                    Node(type=NodeType.BASIC, name=node["name"], label=node["name"])
                )

            nodes.append(Node(type=NodeType.BASIC, name="START", label="START"))

            nodes.append(Node(type=NodeType.BASIC, name="END", label="END"))

            for basic_edge in graph["graph_info"]["basic_edges"]:
                edges.append(
                    Edge(start=basic_edge["start_node"], end=basic_edge["end_node"])
                )

            for conditional_edge in graph["graph_info"]["conditional_edges"]:
                edges.append(
                    Edge(
                        start=conditional_edge["start_node"],
                        end=conditional_edge["end_node"],
                        condition="if else",
                    )
                )

        tools = []

        for predefined_tool in predefined_tools:
            tools.append(
                Node(
                    name=predefined_tool["name"],
                    type=NodeType.TOOL,
                    category=ToolType(predefined_tool["category"]),
                    vulnerabilities=[],  # TODO Add vulnerability mapping
                )
            )

        for custom_tool in custom_tools:
            tools.append(
                Node(
                    name=custom_tool["name"],
                    description=custom_tool["description"],
                    type=NodeType.CUSTOM_TOOL,
                    category=ToolType.DEFAULT,
                    vulnerabilities=[],  # TODO Add vulnerability mapping
                )
            )

        return GraphDefinition(
            name=Path(root_directory).name, nodes=nodes, edges=edges, tools=tools
        )



from .analyze import Analyzer
from .crewai import CrewAIAnalyzer
from .langgraph import LangGraphAnalyzer
from .n8n import N8nAnalyzer

__all__ = ["Analyzer", "LangGraphAnalyzer", "CrewAIAnalyzer", "N8nAnalyzer"]



import json
from importlib import resources
from typing import List, Tuple

from ...graph import (
    EdgeDefinition,
    NodeDefinition,
    NodeType,
)
from .models import N8nConnection, N8nNode


def convert_nodes(n8n_nodes: List[N8nNode]) -> Tuple[List[NodeDefinition], List[NodeDefinition]]:
    nodes = []
    tools = []
    n8n_node_types_file = resources.files(__package__) / "n8n_node_types.json"
    with n8n_node_types_file.open("r") as f:
        n8n_node_types_dict = json.load(f)

    for n8n_node in n8n_nodes:
        type = None
        if n8n_node.type in n8n_node_types_dict:
            type = n8n_node_types_dict.get(n8n_node.type)
        elif n8n_node.type.strip("tool") in n8n_node_types_dict:
            type = n8n_node_types_dict.get(n8n_node.type.strip("tool"))

        if type:
            if type.get("tool_type", False):
                nodes.append(
                    NodeDefinition(
                        type = NodeType.TOOL,
                        name = n8n_node.name,
                        label = n8n_node.name,
                        category = type.get("tool_type")
                    )
                )

                tools.append(
                    NodeDefinition(
                        type = NodeType.TOOL,
                        name = n8n_node.name,
                        label = n8n_node.name,
                        category = type.get("tool_type")
                    )
                )
            else:
                nodes.append(
                    NodeDefinition(
                        type = type.get("node_type"),
                        name = n8n_node.name,
                        label = n8n_node.name
                    )
                )

    return nodes, tools    


def convert_connections(n8n_connections: List[N8nConnection]) -> List[EdgeDefinition]:
    edges = []
    for n8n_connection in n8n_connections:
        edges.append(
            EdgeDefinition(
                start = n8n_connection.start_node,
                end = n8n_connection.end_node
            )
        )

    return edges


import json
from importlib import resources

from agentic_radar.graph import ToolType


def categorize_tool(tool_name: str) -> ToolType:
    """
    Load the tool type from the JSON file and return a ToolType.
    """
    try:
        # Use resources.files().joinpath() to get the path, then use .read_text() to read the file
        json_text = (
            resources.files(__package__).joinpath("tool_categories.json").read_text()
        )
        tool_mapping = json.loads(json_text)
        category = tool_mapping.get(tool_name, "DEFAULT")
        return ToolType[category]
    except (FileNotFoundError, json.JSONDecodeError, KeyError) as e:
        print(f"Error loading tool categories: {e}")
        return ToolType.DEFAULT


import json
import os
from typing import List, Tuple

from ...analysis.analyze import Analyzer
from ...graph import GraphDefinition
from .converter import convert_connections, convert_nodes
from .models import N8nConnection, N8nNode
from .parsing import parse_n8n_connections, parse_n8n_nodes


class N8nAnalyzer(Analyzer):
    def __init__(self):
        super().__init__()

    def parse_n8n_configs(self, root_directory: str) -> Tuple[List[N8nNode], List[N8nConnection], str]:
        n8n_nodes = []
        n8n_connections = []
        workflow_name = ""
        for root, _, files in os.walk(root_directory):
            for file in files:
                if file.endswith(".json"):
                    file_path = os.path.join(root, file)
                    with open(file_path, "r", encoding = "utf-8") as f:
                        config = json.load(f)
                        if config.get("nodes", False) and config.get("connections", False):
                            n8n_nodes.extend(parse_n8n_nodes(config.get("nodes")))
                            n8n_connections.extend(parse_n8n_connections(config.get("connections")))
                            if config.get("name", False):
                                workflow_name = config.get("name")
                            else:
                                workflow_name = file.strip(".json")

        return n8n_nodes, n8n_connections, workflow_name

    def analyze(self, root_directory: str) -> GraphDefinition:
        
        n8n_nodes, n8n_connections, workflow_name = self.parse_n8n_configs(root_directory)

        nodes, tools = convert_nodes(n8n_nodes)
        edges = convert_connections(n8n_connections)

        return GraphDefinition(
            name=workflow_name, nodes=nodes, edges=edges, tools=tools
        )


import pydot


class Edge(pydot.Edge):
    def __init__(self, src: str, dst: str, vulnerable: bool = False):
        super().__init__(src, dst, color="#E20E14" if vulnerable else "#1ED044")


class ConditionalEdge(pydot.Edge):
    def __init__(self, src: str, dst: str, condition: str, vulnerable: bool = False):
        super().__init__(
            src, dst, style="dashed", color="#E20E14" if vulnerable else "#1ED044"
        )
        # self.set_label(condition)  # type: ignore



from typing import Any, Optional

import yaml

from agentic_radar.analysis.crewai.models.tool import CrewAITool
from agentic_radar.analysis.crewai.tool_descriptions import (
    get_crewai_tools_descriptions,
)
from agentic_radar.analysis.utils import walk_yaml_files


def read_yaml_config_file(file_path) -> Any:
    with open(file_path, "r") as file:
        try:
            content = yaml.safe_load(file)
        except Exception as e:
            raise ValueError(f"Cannot parse YAML file: {file_path}. Error: {e}")

        return content


def is_agent_config(file_content: Any) -> bool:
    if not isinstance(file_content, dict):
        return False

    first_value = next(iter(file_content.values()))
    if (
        isinstance(first_value, dict)
        and "role" in first_value
        and "goal" in first_value
        and "backstory" in first_value
    ):
        return True

    return False


def is_task_config(file_content: Any) -> bool:
    if not isinstance(file_content, dict):
        return False

    first_value = next(iter(file_content.values()))
    if (
        isinstance(first_value, dict)
        and "description" in first_value
        and "expected_output" in first_value
    ):
        return True

    return False


def extract_agent_tools_from_config(
    file_content: dict[str, dict[str, Any]], crewai_tool_descriptions: dict[str, str]
) -> dict[str, list[CrewAITool]]:
    assert is_agent_config(file_content)

    yaml_agent_tool_mapping = {}

    for agent in file_content:
        tool_names: list[str] = file_content[agent].get("tools", [])
        tools: list[CrewAITool] = [
            CrewAITool(
                name=tool_name,
                custom=False,
                description=crewai_tool_descriptions.get(tool_name, ""),
            )
            for tool_name in tool_names
        ]
        yaml_agent_tool_mapping[agent] = tools

    return yaml_agent_tool_mapping


def extract_task_agent_from_config(
    file_content: dict[str, dict[str, Any]],
) -> dict[str, str]:
    assert is_task_config(file_content)

    yaml_task_agent_mapping = {}

    for task in file_content:
        agent: Optional[str] = file_content[task].get("agent", None)
        if not agent:
            continue
        yaml_task_agent_mapping[task] = agent

    return yaml_task_agent_mapping


def collect_agent_tools_from_config(
    root_dir: str,
) -> dict[str, dict[str, list[CrewAITool]]]:
    crewai_tools_descriptions = get_crewai_tools_descriptions()
    yaml_file_to_agent_tool_mapping: dict[str, dict[str, list[CrewAITool]]] = {}
    for file in walk_yaml_files(root_dir):
        try:
            file_content = read_yaml_config_file(file)
        except ValueError as e:
            print(f"Error while reading YAML file: {e}")
            continue
        if is_agent_config(file_content):
            yaml_agent_tool_mapping = extract_agent_tools_from_config(
                file_content, crewai_tools_descriptions
            )
            yaml_file_to_agent_tool_mapping[file] = yaml_agent_tool_mapping

    return yaml_file_to_agent_tool_mapping


def collect_task_agents_from_config(root_dir: str) -> dict[str, dict[str, str]]:
    yaml_file_to_task_agent_mapping: dict[str, dict[str, str]] = {}
    for file in walk_yaml_files(root_dir):
        try:
            file_content = read_yaml_config_file(file)
        except ValueError as e:
            print(f"Error while reading YAML file: {e}")
            continue
        if is_task_config(file_content):
            yaml_task_agent_mapping = extract_task_agent_from_config(file_content)
            yaml_file_to_task_agent_mapping[file] = yaml_task_agent_mapping

    return yaml_file_to_task_agent_mapping



from .graph import Graph
from .parse import (
    EdgeDefinition,
    GraphDefinition,
    NodeDefinition,
    from_definition,
    from_dict,
    from_json,
)

__all__ = [
    "Graph",
    "GraphDefinition",
    "NodeDefinition",
    "EdgeDefinition",
    "from_json",
    "from_dict",
    "from_definition",
]



import ast
import json
from typing import Any, Dict, List, Optional, Set, Tuple, Union

from agentic_radar.analysis.utils import walk_python_files_and_notebooks


class GraphInstanceTracker(ast.NodeVisitor):
    """
    A node visitor that:
      1. Finds instantiations of the Graph class (by fully-qualified name).
      2. Tracks method calls on those instances, capturing arguments.
      3. Adds extra logic for 'add_conditional_edges' method:
         - If called with 2 args, the second is the name of a function -> gather return values
         - If called with 3 args, the third is a literal list/dict or a variable referencing one -> gather values
      4. Adds extra logic for "add_node" method:
         - If called with 1 positional argument -> gather information about it
         - If called with 2 positional arguments -> gather information about it
    """

    def __init__(
        self,
        graph_class_fqcn: str,
        command_class_fqn: str,
        add_conditional_edges_method_name: str,
        add_node_method_name: str,
        global_functions,
        global_variables,
    ):
        """
        graph_class_fqcn: The fully qualified class name of the Graph class to look for.
        command_class_fqcn: The fully qualified class name of the Command class to look for.
        add_conditional_edges_method_name: The method name that, when called, triggers addition of conditional edges.
        add_node_method_name: The method name that, when called, triggers addition of a node.
        """
        self.graph_class_fqcn = graph_class_fqcn
        self.command_class_fqn = command_class_fqn
        self.add_conditional_edges_method_name = add_conditional_edges_method_name
        self.add_node_method_name = add_node_method_name

        # Maintain a mapping of local import aliases to fully qualified paths.
        self.import_aliases: Dict[str, str] = {}

        # For "from examples.graph import Graph", store fully qualified references:
        self.import_aliases_fully: Dict[str, str] = {}

        self.instance_methods: Dict[str, Dict[str, List[Dict[str, Any]]]] = {}

        # Track which variables are instances of our target class.
        # If we see "g = Graph()", then self.variable_is_target_instance["g"] = True
        self.variable_is_target_instance: Dict[str, bool] = {}

        # Keep track of all top-level functions
        self.function_defs: Dict[str, Union[ast.FunctionDef, ast.AsyncFunctionDef]] = {}

        # Track variables that hold a literal list or dict, so that if a method is called
        # with that variable, we can retrieve its contents
        self.variable_values: Dict[str, Union[ast.List, ast.Dict]] = {}

        # Tracking global functions and variables that have lists or dictionaries as values for resolving imports
        self.global_functions = global_functions
        self.global_variables = global_variables

    def visit_Import(self, node: ast.Import) -> Any:
        """
        Handle statements like:
          import examples
          import examples.graph as ex
        """
        for alias in node.names:
            local_name = alias.asname if alias.asname else alias.name
            self.import_aliases[local_name] = alias.name

        self.generic_visit(node)

    def visit_ImportFrom(self, node: ast.ImportFrom) -> Any:
        """
        Handle statements like:
          from examples.graph import Graph
          from examples import graph as gph
        """

        if node.module is None:
            self.generic_visit(node)
            return

        base_module = node.module
        for alias in node.names:
            local_name = alias.asname if alias.asname else alias.name
            self.import_aliases_fully[local_name] = f"{base_module}.{alias.name}"

        self.generic_visit(node)

    def visit_FunctionDef(self, node: ast.FunctionDef) -> Any:
        """
        Record the top-level function definitions for later retrieval.
        """
        self.function_defs[node.name] = node
        self.generic_visit(node)

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> Any:
        """
        Same as visit_FunctionDef, but for async functions.
        """
        self.function_defs[node.name] = node
        self.generic_visit(node)

    def visit_Assign(self, node: ast.Assign) -> Any:
        """
        Handle assignments, e.g.:
            g = Graph(...)
            my_list = [1, 2, 3]
            my_dict = {"foo": 42}
        We identify if the assigned value is a call to our target class
        or a list/dict.
        """
        # If there's exactly one target, and it's a Name, e.g., "x = ..."
        if len(node.targets) == 1 and isinstance(node.targets[0], ast.Name):
            var_name = node.targets[0].id

            # If the right side is a Call node, check if it instantiates the Graph class
            if isinstance(node.value, ast.Call):
                if self._call_is_target_class(node.value):
                    self.variable_is_target_instance[var_name] = True

            # If the right side is a list or dict, store it for later.
            if isinstance(node.value, (ast.List, ast.Dict)):
                self.variable_values[var_name] = node.value
            else:
                # If we reassign the variable to something else, remove from variable_values
                if var_name in self.variable_values:
                    del self.variable_values[var_name]

        self.generic_visit(node)

    def visit_Call(self, node: ast.Call) -> Any:
        """
        Resolving method calls on the target Graph class in order to get the arguments with which the methods were called.
        The "add_conditional_edges" and "add_node" methods are examined in more detail.
        """
        # The 'func' of a Call might be an Attribute like g.add_node
        if isinstance(node.func, ast.Attribute):
            # node.func.value might be a Name, e.g., 'g'
            if isinstance(node.func.value, ast.Name):
                instance_name = node.func.value.id

                # Check if that variable is an instance of our target class
                if self.variable_is_target_instance.get(instance_name, False):
                    method_name = node.func.attr

                    # A record of all the arguments of the call
                    call_record = {
                        "positional": [self._stringify_ast_node(a) for a in node.args],
                        "keyword": {
                            kw.arg: self._stringify_ast_node(kw.value)
                            for kw in node.keywords
                            if kw.arg is not None
                        },
                    }

                    if method_name == self.add_conditional_edges_method_name:
                        self._process_add_conditional_edges_method_call(
                            node, call_record
                        )

                    if method_name == self.add_node_method_name:
                        self._handle_add_node_argument(node, call_record)

                    if instance_name not in self.instance_methods:
                        self.instance_methods[instance_name] = {}
                    if method_name not in self.instance_methods[instance_name]:
                        self.instance_methods[instance_name][method_name] = []

                    self.instance_methods[instance_name][method_name].append(
                        call_record
                    )

        self.generic_visit(node)

    # Helpers

    def _handle_add_node_argument(self, call_node: ast.Call, call_record: dict) -> None:
        """
        Resolving the arguments of the "add_node" method in order to find all nodes, along with potential implicitly defined edges.
        """

        num_pos = len(call_node.args)
        if num_pos == 1:
            info, gotos = self._analyze_argument(call_node.args[0])
            call_record["node_definition_argument_info"] = info
            call_record["gotos"] = gotos

        elif num_pos == 2:
            info, gotos = self._analyze_argument(call_node.args[1])
            call_record["node_definition_argument_info"] = info
            call_record["gotos"] = gotos

    def _analyze_argument(
        self, node: ast.AST
    ) -> Tuple[Dict[str, Optional[str]], Set[str]]:
        """
        Return a dictionary describing the argument:
        - "original": how it appears in the code
        - "fq_name": best guess fully qualified name (if we can resolve)
        and all the possible values of the "goto" argument in the node definition if they exist. This is used to determine the implicitly defined edges.
        """
        result: Dict[str, Optional[str]] = {
            "original": self._stringify_ast_node(node),
            "fq_name": None,
        }

        gotos = set()

        # 1) If it's a bare name (like "some_func")
        if isinstance(node, ast.Name):
            fq_name = self._resolve_fq_name(node)
            result["fq_name"] = fq_name

            if fq_name in self.function_defs:
                gotos = self._get_goto_arguments_of_returned_command_objects(
                    self.function_defs[fq_name]
                )

            elif fq_name in self.global_functions:
                gotos = self._get_goto_arguments_of_returned_command_objects(
                    self.global_functions[fq_name]
                )

        # 2) If it's an attribute, e.g. "some_import.func"
        elif isinstance(node, ast.Attribute):
            fq_name = self._resolve_fq_name(node)
            result["fq_name"] = fq_name

            if fq_name in self.function_defs:
                gotos = self._get_goto_arguments_of_returned_command_objects(
                    self.function_defs[fq_name]
                )

            elif fq_name in self.global_functions:
                gotos = self._get_goto_arguments_of_returned_command_objects(
                    self.global_functions[fq_name]
                )

        # 3) If it's a call, e.g. "some_import.function(...)" or "local_func(...)" or "Name()"
        elif isinstance(node, ast.Call):
            func = node.func
            if isinstance(func, (ast.Name, ast.Attribute)):
                fq_name = self._resolve_fq_name(func)
                result["fq_name"] = fq_name

            else:
                result["fq_name"] = result["original"]

        return result, gotos

    def _process_add_conditional_edges_method_call(
        self, call_node: ast.Call, call_record: dict
    ) -> None:
        """
        Resolving the arguments of the "add_conditional_edges" method in order to find edges.
        """

        # Combine all arguments (positional + keyword) in source order.
        # The 'value' of each keyword is the actual expression.
        all_args_in_order = self._get_all_args_in_call(call_node)

        if len(all_args_in_order) == 2:
            second_arg = all_args_in_order[1]
            if isinstance(second_arg, ast.Name):
                func_name = second_arg.id
                if func_name in self.function_defs:
                    returns = self._get_return_expressions(
                        self.function_defs[func_name]
                    )
                    call_record["path"] = {"function_returns": returns}
                elif self._resolve_fq_name(second_arg) in self.global_functions:
                    returns = self._get_return_expressions(
                        self.global_functions[self._resolve_fq_name(second_arg)]
                    )
                    call_record["path"] = {"function_returns": returns}

        elif len(all_args_in_order) == 3:
            third_arg = all_args_in_order[2]

            if isinstance(third_arg, ast.List):
                list_vals = [self._stringify_ast_node(elt) for elt in third_arg.elts]
                call_record["path_map"] = {"list_values": list_vals}

            elif isinstance(third_arg, ast.Dict):
                dict_vals = [self._stringify_ast_node(v) for v in third_arg.values]
                call_record["path_map"] = {"dict_values": dict_vals}

            elif isinstance(third_arg, ast.Name):
                var_name = third_arg.id
                if var_name in self.variable_values:
                    val_node = self.variable_values[var_name]
                    if isinstance(val_node, ast.List):
                        list_vals = [
                            self._stringify_ast_node(elt) for elt in val_node.elts
                        ]
                        call_record["path_map"] = {"list_values": list_vals}
                    elif isinstance(val_node, ast.Dict):
                        dict_vals = [
                            self._stringify_ast_node(v) for v in val_node.values
                        ]
                        call_record["path_map"] = {"dict_values": dict_vals}
                elif self._resolve_fq_name(third_arg) in self.global_variables:
                    val_node = self.global_variables[self._resolve_fq_name(third_arg)]
                    if isinstance(val_node, ast.List):
                        list_vals = [
                            self._stringify_ast_node(elt) for elt in val_node.elts
                        ]
                        call_record["path_map"] = {"list_values": list_vals}
                    elif isinstance(val_node, ast.Dict):
                        dict_vals = [
                            self._stringify_ast_node(v) for v in val_node.values
                        ]
                        call_record["path_map"] = {"dict_values": dict_vals}

    def _call_is_target_class(self, call_node: ast.Call) -> bool:
        """
        Checks if the call instantiates our target Graph class
        """
        fq_name = self._resolve_fq_name(call_node.func)
        return fq_name == self.graph_class_fqcn

    def _resolve_fq_name(self, node) -> str:
        """
        Attempt to resolve a node to a fully qualified name
        """
        if isinstance(node, ast.Name):
            if node.id in self.import_aliases_fully:
                return self.import_aliases_fully[node.id]
            return self.import_aliases.get(node.id, node.id)
        if isinstance(node, ast.Attribute):
            base_fq = self._resolve_fq_name(node.value)
            if base_fq:
                return base_fq + "." + node.attr
            return node.attr
        return ""

    def _stringify_ast_node(self, node) -> str:
        """
        Convert an AST node into a string
        """
        if isinstance(node, ast.Constant):
            if isinstance(node.value, str):
                return node.value
            return repr(node.value)
        if isinstance(node, ast.Name):
            return node.id
        if isinstance(node, ast.Call):
            return f"Call({ast.dump(node)})"
        if isinstance(node, ast.List):
            return (
                "ListLiteral(["
                + ", ".join(self._stringify_ast_node(elt) for elt in node.elts)
                + "])"
            )
        if isinstance(node, ast.Dict):
            keys = [self._stringify_ast_node(k) for k in node.keys]
            vals = [self._stringify_ast_node(v) for v in node.values]
            pairs_str = ", ".join(f"{k}: {v}" for k, v in zip(keys, vals))
            return f"DictLiteral({{{pairs_str}}})"

        return ast.dump(node)

    def _get_all_args_in_call(self, call_node: ast.Call) -> List[ast.expr]:
        """
        Returns a list of all arguments (positional first, then each
        keyword argument's value) in the order they appear in the source.
        """
        combined = []
        combined.extend(call_node.args)

        for kw in call_node.keywords:
            combined.append(kw.value)
        return combined

    def _get_return_expressions(
        self, func_def: Union[ast.FunctionDef, ast.AsyncFunctionDef]
    ) -> List[str]:
        """
        Traverse the body of a function definition and collect all returns
        """

        returns: List[str] = []

        class ReturnCollector(ast.NodeVisitor):
            """
            Inner visitor class that visits Return nodes in the function body.
            We store them in the outer 'returns' list.
            """

            def __init__(inner_self, outer_self):
                super().__init__()
                inner_self.outer = outer_self
                inner_self.variables: Dict[str, List[str]] = {}

            def visit_Return(inner_self, return_node: ast.Return):
                if return_node.value is not None and isinstance(
                    return_node.value, ast.Name
                ):
                    if return_node.value.id in inner_self.variables:
                        returns.extend(inner_self.variables[return_node.value.id])
                    else:
                        returns.append(
                            inner_self.outer._stringify_ast_node(return_node.value)
                        )
                if return_node.value is not None and isinstance(
                    return_node.value, ast.Constant
                ):
                    returns.append(return_node.value.value)

            def visit_Assign(inner_self, node: ast.Assign) -> Any:
                """
                Handle assignments in case code is like:
                if ...:
                    next_node = "node1"
                else:
                    next_node = "node2"

                return next_node
                """
                # If there's exactly one target, and it's a Name, e.g., "x = ..."
                if len(node.targets) == 1 and isinstance(node.targets[0], ast.Name):
                    var_name = node.targets[0].id

                    # If the right side is a literal list or dict, store it for later.
                    if isinstance(node.value, ast.Constant):
                        # and node.value accounts for None
                        if (
                            var_name not in inner_self.variables
                            and node.value.value is not None
                        ):
                            inner_self.variables[var_name] = [node.value.value]
                        elif node.value.value:
                            inner_self.variables[var_name].append(node.value.value)
                    elif isinstance(node.value, ast.Name):
                        # example var1 = var2
                        if node.value.id in inner_self.variables:
                            inner_self.variables[var_name] = inner_self.variables[
                                node.value.id
                            ]
                        else:
                            inner_self.variables[var_name] = [
                                inner_self.outer._stringify_ast_node(node.value)
                            ]

        ReturnCollector(self).visit(func_def)
        return returns

    def _get_goto_arguments_of_returned_command_objects(
        self, func_def: Union[ast.FunctionDef, ast.AsyncFunctionDef]
    ) -> Set[str]:
        """
        Traverse the body of a function definition and collect all possible values of the "goto" argument of a returned Command object
        """

        gotos = set()

        class GotoCollector(ast.NodeVisitor):
            """
            Inner visitor class that visits Return nodes in the function body and sees if a Command object is returned.
            We store the values of the "goto" argument in the outer 'gotos' set.
            """

            def __init__(inner_self, outer_self):
                super().__init__()
                inner_self.outer = outer_self
                inner_self.variables: Dict[str, List[str]] = {}

            def visit_Return(inner_self, return_node: ast.Return):
                if return_node.value is not None and isinstance(
                    return_node.value, ast.Call
                ):
                    call_node = return_node.value
                    # Ends with because we may be using global function definition in which we don't have access to fqn
                    if isinstance(
                        call_node.func, ast.Name
                    ) and inner_self.outer.command_class_fqn.endswith(
                        call_node.func.id
                    ):
                        for kw in call_node.keywords:
                            if kw.arg == "goto":
                                val_node = kw.value
                                if isinstance(val_node, ast.Constant) and isinstance(
                                    val_node.value, str
                                ):
                                    gotos.add(val_node.value)

                                elif isinstance(val_node, ast.List):
                                    for elt in val_node.elts:
                                        if isinstance(elt, ast.Constant) and isinstance(
                                            elt.value, str
                                        ):
                                            gotos.add(elt.value)

                                elif isinstance(val_node, ast.Name):
                                    var_name = val_node.id
                                    if var_name in inner_self.outer.variable_values:
                                        var_node = inner_self.outer.variable_values[
                                            var_name
                                        ]
                                        if isinstance(var_node, ast.List):
                                            list_vals = [
                                                self._stringify_ast_node(elt)
                                                for elt in var_node.elts
                                            ]
                                            for val in list_vals:
                                                gotos.add(val)
                                    elif var_name in inner_self.variables:
                                        list_vals = inner_self.variables[var_name]
                                        for val in list_vals:
                                            gotos.add(val)
                                    else:
                                        gotos.add(
                                            inner_self.outer._stringify_ast_node(
                                                val_node
                                            )
                                        )

            def visit_Assign(inner_self, node: ast.Assign) -> Any:
                """
                Handle assignments in case code is like:
                if ...:
                    next_node = "node1"
                else:
                    next_node = "node2"

                return Command(...goto=next_node)
                """
                # If there's exactly one target, and it's a Name, e.g., "x = ..."
                if len(node.targets) == 1 and isinstance(node.targets[0], ast.Name):
                    var_name = node.targets[0].id

                    # If the right side is a literal list or dict, store it for later.
                    if isinstance(node.value, ast.Constant):
                        # and node.value accounts for None
                        if (
                            var_name not in inner_self.variables
                            and node.value.value is not None
                        ):
                            inner_self.variables[var_name] = [node.value.value]
                        elif node.value.value:
                            inner_self.variables[var_name].append(node.value.value)
                    elif isinstance(node.value, ast.Name):
                        # example var1 = var2
                        if node.value.id in inner_self.variables:
                            inner_self.variables[var_name] = inner_self.variables[
                                node.value.id
                            ]
                        else:
                            inner_self.variables[var_name] = [
                                inner_self.outer._stringify_ast_node(node.value)
                            ]
                    elif isinstance(node.value, ast.List):
                        if var_name not in inner_self.variables and node.value:
                            inner_self.variables[var_name] = [
                                self._stringify_ast_node(elt) for elt in node.value.elts
                            ]
                        elif node.value:
                            inner_self.variables[var_name].extend(
                                [
                                    self._stringify_ast_node(elt)
                                    for elt in node.value.elts
                                ]
                            )

        GotoCollector(self).visit(func_def)
        return gotos


def parse_python_file(
    filepath: str,
    target_class: str,
    command_class_fqn: str,
    add_conditional_edges_method_name: str,
    add_node_method_name: str,
    global_functions,
    global_variables,
) -> Dict[str, Dict[str, List[Dict[str, Any]]]]:
    """
    Parse a single Python file for:
      - Instantiations of the Graph class
      - Method calls on those instances
    """

    source = ""
    if filepath.endswith(".py"):
        with open(filepath, "r", encoding="utf-8") as f:
            source = f.read()
    elif filepath.endswith(".ipynb"):
        with open(filepath, "r", encoding="utf-8") as f:
            notebook = json.load(f)
            for cell in notebook["cells"]:
                if cell["cell_type"] == "code":
                    for row in cell["source"]:
                        source += row

    try:
        tree = ast.parse(source, filename=filepath)
    except Exception as e:
        print(f"Cannot parse Python module: {filepath}. Error: {e}")

    tracker = GraphInstanceTracker(
        target_class,
        command_class_fqn,
        add_conditional_edges_method_name,
        add_node_method_name,
        global_functions,
        global_variables,
    )

    tracker.visit(tree)

    return tracker.instance_methods


def walk_directory_and_parse(
    root_dir: str,
    target_class: str,
    command_class_fqn: str,
    add_conditional_edges_method_name: str,
    add_node_method_name: str,
    global_functions,
    global_variables,
) -> Dict[str, Dict[str, List[Dict[str, Any]]]]:
    """
    Recursively walk a directory (root_dir), parse each .py and .ipynb file,
    and merge results for the target Graph class
    """
    combined_results: Dict[str, Dict[str, List[Dict[str, Any]]]] = {}

    for file_path in walk_python_files_and_notebooks(root_dir):
        file_results = parse_python_file(
            file_path,
            target_class,
            command_class_fqn,
            add_conditional_edges_method_name,
            add_node_method_name,
            global_functions,
            global_variables,
        )

        for instance_name, methods_dict in file_results.items():
            if instance_name not in combined_results:
                combined_results[instance_name] = {}
            for method_name, calls in methods_dict.items():
                if method_name not in combined_results[instance_name]:
                    combined_results[instance_name][method_name] = []
                combined_results[instance_name][method_name].extend(calls)

    return combined_results


def parse_all_graph_instances_in_directory(
    root_directory,
    graph_class_fqcn,
    command_class_fqn,
    add_conditional_edges_method_name,
    add_node_method_name,
    global_functions,
    global_variables,
):
    results = walk_directory_and_parse(
        root_directory,
        graph_class_fqcn,
        command_class_fqn,
        add_conditional_edges_method_name,
        add_node_method_name,
        global_functions,
        global_variables,
    )
    graphs = []
    for graph, call_records in results.items():
        nodes = []
        basic_edges = []
        conditional_edges = []

        if call_records.get("add_node", False):
            call_data = call_records.get("add_node")
            for single_call in call_data:
                # The actual name if the method received 2 arguments, and the name of the function if it received one
                node_name = single_call["positional"][0]
                node_definition = single_call["node_definition_argument_info"]
                nodes.append({"name": node_name, "definition": node_definition})

            # Resolving the goto arguments of the Command objects
            for single_call in call_data:
                node_name = single_call["positional"][0]
                all_node_names = [node["name"] for node in nodes] + ["START", "END"]
                resolved_gotos = [
                    goto for goto in single_call["gotos"] if goto in all_node_names
                ]

                if len(resolved_gotos) > 1:
                    for resolved_goto in resolved_gotos:
                        conditional_edges.append(
                            {
                                "resolved": True,
                                "start_node": node_name,
                                "end_node": resolved_goto,
                            }
                        )
                elif len(resolved_gotos) == 1:
                    basic_edges.append(
                        {"start_node": node_name, "end_node": resolved_gotos[0]}
                    )

        if call_records.get("add_edge", False):
            all_node_names = [node["name"] for node in nodes] + ["START", "END"]
            call_data = call_records.get("add_edge")
            for single_call in call_data:
                nodes_in_edge = []
                # Since the call always contains two node names, we just add names from positional and keyword arguments (in this exact order)
                for node_name in single_call["positional"]:
                    nodes_in_edge.append(node_name)
                for _, node_name in single_call["keyword"].items():
                    nodes_in_edge.append(node_name)

                if (
                    nodes_in_edge[0] in all_node_names
                    and nodes_in_edge[1] in all_node_names
                ):
                    basic_edges.append(
                        {"start_node": nodes_in_edge[0], "end_node": nodes_in_edge[1]}
                    )

        if call_records.get("add_conditional_edges", False):
            all_node_names = [node["name"] for node in nodes] + ["START", "END"]
            call_data = call_records.get("add_conditional_edges")
            for single_call in call_data:
                # Resolving arguments to find the source node
                arguments = []
                for argument in single_call["positional"]:
                    arguments.append(argument)
                for _, argument in single_call["keyword"].items():
                    arguments.append(argument)

                if single_call.get("path", False):
                    if single_call["path"].get("function_returns", False):
                        for end_node in single_call["path"].get("function_returns"):
                            if (
                                arguments[0] in all_node_names
                                and end_node in all_node_names
                            ):
                                conditional_edges.append(
                                    {
                                        "resolved": True,
                                        "start_node": arguments[0],
                                        "end_node": end_node,
                                    }
                                )
                elif single_call.get("path_map", False):
                    if single_call["path_map"].get("list_values", False):
                        for end_node in single_call["path_map"].get("list_values"):
                            if (
                                arguments[0] in all_node_names
                                and end_node in all_node_names
                            ):
                                conditional_edges.append(
                                    {
                                        "resolved": True,
                                        "start_node": arguments[0],
                                        "end_node": end_node,
                                    }
                                )
                    elif single_call["path_map"].get("dict_values", False):
                        for end_node in single_call["path_map"].get("dict_values"):
                            if (
                                arguments[0] in all_node_names
                                and end_node in all_node_names
                            ):
                                conditional_edges.append(
                                    {
                                        "resolved": True,
                                        "start_node": arguments[0],
                                        "end_node": end_node,
                                    }
                                )
        if call_records.get("set_entry_point", False):
            call_data = call_records.get("set_entry_point")
            entrypoints = []
            for single_call in call_data:
                for node_name in single_call["positional"]:
                    entrypoints.append(node_name)
                for _, node_name in single_call["keyword"].items():
                    entrypoints.append(node_name)

            for node_name in entrypoints:
                basic_edges.append({"start_node": "START", "end_node": node_name})
        graphs.append(
            {
                "graph_name": graph,
                "graph_info": {
                    "nodes": nodes,
                    "basic_edges": basic_edges,
                    "conditional_edges": conditional_edges,
                },
            }
        )
    return graphs



from agentic_radar.analysis.crewai.models import CrewAIGraph, CrewAINodeType
from agentic_radar.analysis.crewai.tool_categorizer import categorize_tool
from agentic_radar.graph import (
    EdgeDefinition,
    GraphDefinition,
    NodeDefinition,
    NodeType,
    ToolType,
)


def convert_graph(crewai_graph: CrewAIGraph) -> GraphDefinition:
    """Convert CrewAI Graph to GraphDefinition."""
    nodes = []
    edges = []
    tools = []

    for node in crewai_graph.nodes:
        if node.type == CrewAINodeType.AGENT:
            node_type = NodeType.AGENT
        elif node.type == CrewAINodeType.TOOL:
            node_type = NodeType.TOOL
        elif node.type == CrewAINodeType.CUSTOM_TOOL:
            node_type = NodeType.CUSTOM_TOOL
        elif node.type == CrewAINodeType.BASIC:
            node_type = NodeType.BASIC
        else:
            node_type = NodeType.DEFAULT

        category = None
        if node_type == NodeType.TOOL:
            category = categorize_tool(node.name)
        elif node_type == NodeType.CUSTOM_TOOL:
            category = ToolType.DEFAULT

        output_node = NodeDefinition(
            type=node_type,
            name=node.name,
            label=node.name,
            category=category,
            description=node.description,
        )

        nodes.append(output_node.model_copy(deep=True))

        if node.type == CrewAINodeType.TOOL or node.type == CrewAINodeType.CUSTOM_TOOL:
            tools.append(output_node.model_copy(deep=True))

    for edge in crewai_graph.edges:
        edges.append(
            EdgeDefinition(
                start=edge.start_node,
                end=edge.end_node,
                condition=edge.condition,
            )
        )

    return GraphDefinition(
        name=crewai_graph.name, nodes=nodes, edges=edges, tools=tools
    )



import ast

from agentic_radar.analysis.crewai.models.tool import CrewAITool
from agentic_radar.analysis.crewai.tool_descriptions import (
    get_crewai_tools_descriptions,
)
from agentic_radar.analysis.utils import walk_python_files


class PredefinedToolsVisitor(ast.NodeVisitor):
    CREWAI_TOOLS_MODULE = "crewai_tools"

    def __init__(self):
        self.known_tool_aliases: set[str] = set()
        self.predefined_tools: dict[str, str] = {}

    def visit_ImportFrom(self, node):
        """Track imports like 'from crewai_tools import FileReadTool'."""
        if node.module == self.CREWAI_TOOLS_MODULE:
            for alias in node.names:
                tool_name = alias.asname or alias.name
                self.known_tool_aliases.add(tool_name)

    def visit_Assign(self, node):
        """Tracks assigments like some_tool = FileReadTool()"""
        for target in node.targets:
            if not isinstance(target, ast.Name):
                continue

            if not isinstance(node.value, ast.Call):
                continue

            if (
                isinstance(node.value.func, ast.Name)
                and node.value.func.id in self.known_tool_aliases
            ):
                self.predefined_tools[target.id] = node.value.func.id
            elif isinstance(node.value, ast.Attribute):
                if (
                    self.CREWAI_TOOLS_MODULE in node.value
                    and node.value.attr in self.known_tool_aliases
                ):
                    self.predefined_tools[target.id] = node.value.attr


def collect_predefined_tools(root_dir: str) -> tuple[set, dict]:
    """Parses all Python modules in the given directory and collects predefined tools.

    Args:
        root_dir (str): Path to the codebase directory

    Returns:
        tuple[set[str], dict[str, CrewAITool]]: A tuple containing a set of known tool aliases and a dictionary of predefined tool variables
    """
    known_tool_aliases: set[str] = set()
    predefined_tools: dict[str, str] = {}

    for file in walk_python_files(root_dir):
        with open(file, "r") as f:
            try:
                tree = ast.parse(f.read())
            except Exception as e:
                print(f"Cannot parse Python module: {file}. Error: {e}")
                continue
            predefined_tools_visitor = PredefinedToolsVisitor()
            predefined_tools_visitor.visit(tree)
            known_tool_aliases |= predefined_tools_visitor.known_tool_aliases
            predefined_tools |= predefined_tools_visitor.predefined_tools

    # Add descriptions and wrap tools inside CrewAITool instances
    tools_descriptions = get_crewai_tools_descriptions()
    predefined_tools_with_descriptions = {
        var_name: CrewAITool(
            name=tool_name,
            custom=False,
            description=tools_descriptions.get(tool_name, ""),
        )
        for var_name, tool_name in predefined_tools.items()
    }

    return known_tool_aliases, predefined_tools_with_descriptions



import ast
import json
from typing import Dict, List

from agentic_radar.analysis.utils import walk_python_files_and_notebooks


def extract_custom_tools_with_ast(
    file_content: str, file_path: str
) -> List[Dict[str, str]]:
    custom_tools = []

    try:
        tree = ast.parse(file_content)
    except Exception as e:
        print(f"Cannot parse Python module: {file_path}. Error: {e}")

    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef):
            decorator_names = []
            for decorator in node.decorator_list:
                if isinstance(decorator, ast.Name):
                    decorator_names.append(decorator.id)
                elif isinstance(decorator, ast.Call):
                    if isinstance(decorator.func, ast.Name):
                        decorator_names.append(decorator.func.id)
                    elif isinstance(decorator.func, ast.Attribute):
                        decorator_names.append(decorator.func.attr)

            if "tool" in decorator_names:
                custom_tools.append(
                    {
                        "name": node.name,
                        "filepath": file_path,
                        "description": ast.get_docstring(node) or "",
                    }
                )

    return custom_tools


def get_all_custom_tools_from_directory(directory_path: str) -> List[Dict[str, str]]:
    all_custom_tools = []
    for file_path in walk_python_files_and_notebooks(directory_path):
        if file_path.endswith(".py"):
            with open(file_path, "r", encoding="utf-8") as f:
                file_content = f.read()
                custom_tools = extract_custom_tools_with_ast(
                    file_content, file_path
                )
                for single_custom_tool in custom_tools:
                    all_custom_tools.append(single_custom_tool)
        elif file_path.endswith(".ipynb"):
            with open(file_path, "r", encoding="utf-8") as f:
                notebook = json.load(f)
                file_content = ""
                for cell in notebook["cells"]:
                    if cell["cell_type"] == "code":
                        for row in cell["source"]:
                            file_content += row
                custom_tools = extract_custom_tools_with_ast(
                    file_content, file_path
                )
                for single_custom_tool in custom_tools:
                    all_custom_tools.append(single_custom_tool)

    return all_custom_tools



from .graph import (
    EdgeDefinition,
    GraphDefinition,
    NodeDefinition,
)
from .report import generate

__all__ = [
    "EdgeDefinition",
    "GraphDefinition",
    "NodeDefinition",
    "generate",
]



import datetime
import os
import time
from typing import Optional

import typer
from pydantic import BaseModel
from typing_extensions import Annotated

from agentic_radar import __version__
from agentic_radar.analysis import (
    Analyzer,
    CrewAIAnalyzer,
    LangGraphAnalyzer,
    N8nAnalyzer,
)
from agentic_radar.mapper import map_vulnerabilities
from agentic_radar.report import (
    EdgeDefinition,
    GraphDefinition,
    NodeDefinition,
    generate,
)
from agentic_radar.utils import sanitize_graph


class Args(BaseModel):
    input_directory: str
    output_file: str
    version: Optional[bool]


app = typer.Typer()
args: Args


def version_callback(value: bool):
    if value:
        print(f"SplxAI Agentic Radar Version: {__version__}")
        raise typer.Exit()


@app.callback()
def _main(
    input_directory: Annotated[
        str,
        typer.Option(
            "--input-dir",
            "-i",
            help="Path to the directory where all the code is",
            envvar="AGENTIC_RADAR_INPUT_DIRECTORY",
        ),
    ] = ".",
    output_file: Annotated[
        str,
        typer.Option(
            "--output-file",
            "-o",
            help="Where should the output report be stored",
            envvar="AGENTIC_RADAR_OUTPUT_FILE",
        ),
    ] = f"report_{datetime.datetime.fromtimestamp(time.time()).strftime('%Y%m%d_%H%M%S')}.html",
    version: Annotated[
        Optional[bool],
        typer.Option("--version", callback=version_callback, is_eager=True),
    ] = None,
):
    global args

    if not os.path.isdir(input_directory):
        print(f"Input directory '{input_directory}' does not exist.")
        raise typer.Exit(code=1)
    
    args = Args(
        input_directory=input_directory, output_file=output_file, version=version
    )


def analyze_and_generate_report(framework: str, analyzer: Analyzer):
    print(f"Analyzing {args.input_directory} for {framework} graphs")
    graph = analyzer.analyze(args.input_directory)

    if len(graph.nodes) <= 2: # Only start and end nodes are present
        print(f"Agentic Radar didn't find any agentic workflow in input directory: {args.input_directory}")
        raise typer.Exit(code=1)
    
    sanitize_graph(graph)
    
    print("Mapping vulnerabilities")
    map_vulnerabilities(graph)
    pydot_graph = GraphDefinition(
        framework=framework,
        name=graph.name,
        nodes=[
            NodeDefinition.model_validate(n, from_attributes=True) for n in graph.nodes
        ],
        edges=[
            EdgeDefinition.model_validate(e, from_attributes=True) for e in graph.edges
        ],
        tools=[
            NodeDefinition.model_validate(t, from_attributes=True) for t in graph.tools
        ],
    )
    print("Generating report")
    generate(pydot_graph, args.output_file)
    print(f"Report {args.output_file} generated")


@app.command("langgraph", help="Scan code written with LangGraph")
def langgraph():
    analyze_and_generate_report("LangGraph", LangGraphAnalyzer())


@app.command("crewai", help="Scan code written with CrewAI")
def crewai():
    analyze_and_generate_report("CrewAI", CrewAIAnalyzer())

@app.command("n8n", help="Scan a n8n workflow configuration JSON")
def n8n():
    analyze_and_generate_report("n8n", N8nAnalyzer())


if __name__ == "__main__":
    app()



import ast
from typing import Optional

from agentic_radar.analysis.crewai.parsing.utils import (
    find_return_of_function_call,
    is_function_call,
)
from agentic_radar.analysis.crewai.parsing.yaml_config import (
    collect_task_agents_from_config,
)
from agentic_radar.analysis.utils import walk_python_files


class TasksVisitor(ast.NodeVisitor):
    CREWAI_TASK_CLASS = "Task"

    def __init__(self, agents: set[str]):
        self.agents = agents
        self.yaml_config_paths: list[str] = []
        self.task_agent_mapping: dict[str, str] = {}

    def _find_task_return(self, node: ast.AST) -> Optional[ast.Call]:
        """
        Recursively search for and return the Task constructor node from a return statement.

        Args:
            node: An AST node to search within

        Returns:
            The Task constructor node if found, otherwise None
        """
        return find_return_of_function_call(node, self.CREWAI_TASK_CLASS)

    def _is_task_constructor(self, node) -> bool:
        """Check if a node is a call to Task()"""
        return is_function_call(node, self.CREWAI_TASK_CLASS)

    def _extract_task_agent(self, task_node: ast.Call) -> Optional[str]:
        """
        Extract the agent name from a Task constructor node.

        Args:
            task_node: The Task constructor AST node

        Returns:
            The agent name used by this task if found, otherwise None
        """

        def get_agent_name(agent_node: ast.AST) -> Optional[str]:
            if isinstance(agent_node, ast.Name):
                return agent_node.id
            elif isinstance(agent_node, ast.Call) and isinstance(
                agent_node.func, (ast.Name, ast.Attribute)
            ):
                return (
                    agent_node.func.id
                    if isinstance(agent_node.func, ast.Name)
                    else agent_node.func.attr
                )
            return None

        for keyword in task_node.keywords:
            if keyword.arg == "agent":
                agent_node = keyword.value
                agent_name = get_agent_name(agent_node)
                if agent_name and agent_name in self.agents:
                    return agent_name
                print(f"Unknown or unexpected agent {agent_node} for task {task_node}")
                return None
        return None

    def _handle_string_node(self, s: str) -> None:
        if not s.endswith(".yaml") and not s.endswith(".yml"):
            return

        self.yaml_config_paths.append(s)

    def visit_FunctionDef(self, node):
        """Track functions that return a Task instance."""

        task_node = self._find_task_return(node.body)
        if not task_node:
            self.generic_visit(node)
            return

        agent = self._extract_task_agent(task_node)
        if not agent:
            print(f"Task {node.name} does not specify an agent")
            self.generic_visit(node)
            return

        self.task_agent_mapping[node.name] = agent

        self.generic_visit(node)

    def visit_Assign(self, node):
        """Track assignments that return a Task instance."""
        if not isinstance(node.value, ast.Call):
            self.generic_visit(node)
            return

        if not self._is_task_constructor(node.value):
            self.generic_visit(node)
            return

        # Handles cases like task = Task(...)
        for target in node.targets:
            if isinstance(target, ast.Name):
                agent = self._extract_task_agent(node.value)
                if not agent:
                    print(f"Task {target.id} does not specify an agent")
                    return
                self.task_agent_mapping[target.id] = agent

        self.generic_visit(node)

    def visit_Constant(self, node):  # To visit strings in Python >= 3.8
        """Tracks strings that represent path to yaml configuration files."""
        if isinstance(node.value, str):
            self._handle_string_node(node.value)

    def visit_Str(self, node):  # To visit strings in Python < 3.8
        """Tracks strings that represent path to yaml configuration files."""
        self._handle_string_node(node.s)


def collect_tasks(root_dir: str, agents: set[str]) -> dict[str, str]:
    """Parses all Python modules in the given directory and collects task-agent mappings.

    Args:
        root_dir (str): Path to the codebase directory
        agents (set[str]): Set of all known agent names

    Returns:
        dict[str, str]: A dictionary mapping task names to agent names
    """
    task_agent_mapping: dict[str, str] = {}

    yaml_file_to_task_agents_mapping = collect_task_agents_from_config(root_dir)

    for file in walk_python_files(root_dir):
        with open(file, "r") as f:
            try:
                tree = ast.parse(f.read())
            except Exception as e:
                print(f"Cannot parse Python module: {file}. Error: {e}")
                continue
            tasks_visitor = TasksVisitor(agents=agents)
            tasks_visitor.visit(tree)
            task_agent_mapping |= tasks_visitor.task_agent_mapping

            # Add task-agent mapping from YAML config files
            for found_config_path in tasks_visitor.yaml_config_paths:
                for (
                    config_path,
                    yaml_task_agent_mapping,
                ) in yaml_file_to_task_agents_mapping.items():
                    if found_config_path not in config_path:
                        continue
                    task_agent_mapping = yaml_task_agent_mapping | task_agent_mapping

    return task_agent_mapping



import ast
import importlib.util
import os
import re
from typing import Optional


def is_package_installed(package_name: str) -> bool:
    """
    Check if a package is installed in the current environment.

    Args:
        package_name: Name of the package to check

    Returns:
        True if the package is installed, False otherwise
    """
    spec = importlib.util.find_spec(package_name)
    return spec is not None


def parse_init_imports(init_path: str) -> list[tuple[str, str]]:
    """
    Parse the import statements from __init__.py to extract tool classes and their module paths.

    Args:
        init_path: Path to the __init__.py file

    Returns:
        List of tuples containing (class_name, module_path)
    """
    try:
        with open(init_path, "r", encoding="utf-8") as file:
            content = file.read()

        # Parse the file
        parsed_ast = ast.parse(content)

        imports = []

        # Look for import statements
        for node in parsed_ast.body:
            if isinstance(node, ast.ImportFrom):
                # Get the module path (e.g., '.ai_mind_tool.ai_mind_tool')
                module_path = node.module

                if not module_path:
                    continue

                # Get the imports (e.g., 'AIMindTool')
                for name in node.names:
                    imports.append((name.name, module_path))

        return imports

    except Exception as e:
        print(f"Error parsing imports from {init_path}: {e}")
        return []


def find_tool_directory(tools_dir: str, module_path: str) -> Optional[str]:
    """
    Find the directory containing the tool based on its module path.

    Args:
        tools_dir: Base directory of all tools
        module_path: Relative module path (e.g., '.ai_mind_tool.ai_mind_tool')

    Returns:
        Path to the tool directory if found, None otherwise
    """
    # Remove leading dot if present
    if module_path.startswith("."):
        module_path = module_path[1:]

    # Split module path into parts
    parts = module_path.split(".")

    # Try to find the tool directory
    if len(parts) >= 1:
        # Try direct match (most tools have their own directory)
        tool_dir = os.path.join(tools_dir, parts[0])
        if os.path.isdir(tool_dir):
            return tool_dir

    return None


def extract_readme_content(readme_path: str) -> Optional[str]:
    """
    Extract content from a README.md file.

    Args:
        readme_path: Path to the README.md file

    Returns:
        Content of the README.md file if found, None otherwise
    """
    try:
        if os.path.exists(readme_path):
            with open(readme_path, "r", encoding="utf-8") as file:
                return file.read()
        return None
    except Exception as e:
        print(f"Error reading README file at {readme_path}: {e}")
        return None


def extract_description_from_readme(readme_content: str) -> Optional[str]:
    """
    Extract the description section from a README.md file.

    Args:
        readme_content: Content of the README.md file

    Returns:
        Description section if found, None otherwise
    """
    if not readme_content:
        return None

    # Try to find description section
    description_match = re.search(
        r"## Description\s+(.*?)(?=##|\Z)", readme_content, re.DOTALL
    )
    if description_match:
        return description_match.group(1).strip()

    # If no description section found, try to use the content after the title
    title_match = re.search(r"# (.*?)\s+(.*?)(?=##|\Z)", readme_content, re.DOTALL)
    if title_match:
        return title_match.group(2).strip()

    # If still nothing found, return first paragraph
    first_para_match = re.search(r"^(.*?)(?=\n\n|\Z)", readme_content, re.DOTALL)
    if first_para_match:
        return first_para_match.group(1).strip()

    return None


def get_crewai_tools_descriptions() -> dict[str, str]:
    """
    Extract tool descriptions from README.md files in the tool directories.

    Returns:
        Dictionary mapping tool names to their descriptions
    """
    # First check if crewai_tools is installed
    if not is_package_installed("crewai_tools"):
        print("crewai_tools is not installed in the current environment.")
        return {}

    # Find the package location
    spec = importlib.util.find_spec("crewai_tools")
    if not spec or not spec.submodule_search_locations:
        return {}

    package_dir = spec.submodule_search_locations[0]
    tools_dir = os.path.join(package_dir, "tools")
    init_file = os.path.join(tools_dir, "__init__.py")

    if not os.path.exists(init_file):
        print(f"__init__.py not found at {init_file}")
        return {}

    # Parse imports from __init__.py
    imports = parse_init_imports(init_file)

    descriptions = {}

    for class_name, module_path in imports:
        # Find the tool directory
        tool_dir = find_tool_directory(tools_dir, module_path)

        if not tool_dir:
            descriptions[class_name] = f"Tool directory not found for {module_path}"
            continue

        # Look for README.md in the tool directory
        readme_path = os.path.join(tool_dir, "README.md")
        readme_content = extract_readme_content(readme_path)

        if readme_content:
            description = extract_description_from_readme(readme_content)
            if description:
                descriptions[class_name] = description
            else:
                descriptions[class_name] = "No description found in README.md"
        else:
            # Try to find README.md by traversing subdirectories
            readme_found = False
            for root, _, files in os.walk(tool_dir):
                if "README.md" in files:
                    readme_path = os.path.join(root, "README.md")
                    readme_content = extract_readme_content(readme_path)
                    if readme_content:
                        description = extract_description_from_readme(readme_content)
                        if description:
                            descriptions[class_name] = description
                            readme_found = True
                            break

            if not readme_found:
                descriptions[class_name] = "README.md not found for this tool"

    return descriptions



from importlib import resources
from typing import Optional

import pydot

from ...graph import ToolType


class AgentNode(pydot.Node):
    def __init__(self, name: str, label: str):
        super().__init__(
            name,
            label="",
            xlabel=f"< <B>{label}</B> >",
            image=str(resources.files(__package__) / "assets" / "agent.svg"),
            imagescale=True,
            shape="circle",
            color="#164FAE",
            penwidth=20,
            fillcolor="#164FAE",
            labelloc="b",
            style="rounded,filled",
            regular=True,
            width=1.2,
            fixedsize=True,
        )


class BasicNode(pydot.Node):
    def __init__(self, name: str, label: str):
        super().__init__(
            name,
            label="",
            xlabel=f"< <B>{label}</B> >",
            image=str(resources.files(__package__) / "assets" / "basic.svg"),
            imagescale=True,
            fixedsize=True,
            shape="diamond",
            color="none",
            fillcolor="#039A93",
            labelloc="b",
            style="rounded,filled",
            regular=True,
            width=1.2,
        )


class ToolNode(pydot.Node):
    def __init__(self, name: str, label: str, tool_type: Optional[ToolType] = None):
        ttype = str(tool_type or ToolType.DEFAULT)
        super().__init__(
            name,
            label="",
            xlabel=f"< <B>{label}</B><BR/>{ttype} >",
            image=str(
                resources.files(__package__) / "assets" / "tools" / f"{ttype}.svg"
            ),
            imagescale=True,
            fixedsize=True,
            shape="box",
            color="none",
            fillcolor="#A937D6",
            style="rounded,filled",
            labelloc="b",
            regular=True,
            width=1,
        )


class CustomToolNode(pydot.Node):
    def __init__(self, name: str, label: str):
        super().__init__(
            name,
            label="",
            xlabel=f"< <B>{label}</B> >",
            image=str(resources.files(__package__) / "assets" / "custom_tool.svg"),
            imagescale=True,
            shape="box",
            color="#A937D6",
            fillcolor="#F7DEFC",
            style="rounded,filled,dashed",
            labelloc="b",
            regular=True,
            width=1,
            fixedsize=True,
        )



from typing import List

import pydot
from pydantic import BaseModel, Field

from ... import graph
from .edge import ConditionalEdge, Edge
from .graph import Graph
from .node import (
    AgentNode,
    BasicNode,
    CustomToolNode,
    ToolNode,
)


class NodeDefinition(graph.NodeDefinition):
    def to_pydot(self) -> pydot.Node:
        if self.node_type == graph.NodeType.AGENT:
            return AgentNode(self.name, self.label or self.name)
        if self.node_type == graph.NodeType.BASIC:
            return BasicNode(self.name, self.label or self.name)
        if self.node_type == graph.NodeType.TOOL:
            return ToolNode(self.name, self.label or self.name, self.category)
        if self.node_type == graph.NodeType.CUSTOM_TOOL:
            return CustomToolNode(self.name, self.label or self.name)
        raise ValueError(f"Unknown node type: {self.node_type}")


class EdgeDefinition(graph.EdgeDefinition):
    def to_pydot(self) -> pydot.Edge:
        if self.condition is not None:
            return ConditionalEdge(self.start, self.end, self.condition)
        return Edge(self.start, self.end)


class GraphDefinition(BaseModel):
    name: str
    framework: str

    nodes: List[NodeDefinition]
    edges: List[EdgeDefinition]

    tools: List[NodeDefinition] = Field(default_factory=list)


def from_definition(definition: GraphDefinition) -> Graph:
    g = Graph(
        nodes=[node.to_pydot() for node in definition.nodes],
        edges=[edge.to_pydot() for edge in definition.edges],
    )
    return g


def from_json(definition: str) -> Graph:
    return from_definition(GraphDefinition.model_validate_json(definition))


def from_dict(definition: dict) -> Graph:
    return from_definition(GraphDefinition.model_validate(definition))



import datetime
import os
import re
from typing import Dict, List, Optional

import jinja2
from pydantic import BaseModel, Field

from agentic_radar import __version__

from ..graph import NodeType
from .graph import (
    GraphDefinition,
    from_definition,
)


class Vulnerability(BaseModel):
    name: str
    description: str
    security_framework_mapping: Dict[str, str]
    remediation: str


class Tool(BaseModel):
    name: str
    category: str
    description: Optional[str] = None
    vulnerabilities: List[Vulnerability] = Field(default_factory=list)


class ReportData(BaseModel):
    project_name: str
    framework: str
    timestamp: str
    graph: str

    count: Dict[str, int]

    tools: List[Tool]
    scanner_version: str


def generate(graph: GraphDefinition, out_file: str):
    svg = from_definition(graph).generate()

    # remove xml definition, and put relative width and height
    svg_lines = svg.splitlines()[1:]
    svg_lines[0] = re.sub(r'width="(.*?)"', 'width="100%"', svg_lines[0])
    svg_lines[0] = re.sub(
        r'height="(.*?)"', 'height="100%" preserveAspectRatio="xMidYMin"', svg_lines[0]
    )
    svg = "\n".join(svg_lines)

    env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(
            os.path.join(os.path.dirname(__file__), "templates")
        )
    )

    template = env.get_template("template.html.jinja")
    tools = [Tool.model_validate(t, from_attributes=True) for t in graph.tools]
    template.stream(
        **ReportData(
            project_name=graph.name,
            framework=graph.framework,
            timestamp=datetime.datetime.now().strftime("%x %X"),
            graph=svg,
            count={
                "tools": len(tools),
                "agents": len(
                    [x for x in graph.nodes if x.node_type == NodeType.AGENT]
                ),
                "vulnerability": len(
                    [v for t in graph.tools for v in t.vulnerabilities]
                ),
            },
            tools=tools,
            scanner_version=__version__,
        ).model_dump()
    ).dump(out_file)



from typing import Optional

from pydantic import BaseModel, Field


class N8nNode(BaseModel):
    type: str = Field(..., description="Type of the node")
    name: str = Field(..., description="Name of the node")
    id: str = Field(..., description="Id of the node")


class N8nConnection(BaseModel):
    start_node: str = Field(..., description="Id of the starting node")
    end_node: str = Field(..., description="Id of the end node")
    condition: Optional[str] = Field(
        default=None, description="Optional condition for the edge"
    )


import ast
from typing import Optional

from agentic_radar.analysis.crewai.models.tool import CrewAITool
from agentic_radar.analysis.crewai.parsing.utils import (
    find_return_of_function_call,
    is_function_call,
)
from agentic_radar.analysis.crewai.parsing.yaml_config import (
    collect_agent_tools_from_config,
)
from agentic_radar.analysis.crewai.tool_descriptions import (
    get_crewai_tools_descriptions,
)
from agentic_radar.analysis.utils import walk_python_files


class AgentsVisitor(ast.NodeVisitor):
    CREWAI_AGENT_CLASS = "Agent"

    def __init__(
        self,
        known_tool_aliases: set[str],
        predefined_tools: dict[str, CrewAITool],
        custom_tools: dict[str, CrewAITool],
    ):
        self.known_tool_aliases = known_tool_aliases
        self.predefined_tools = predefined_tools
        self.custom_tools = custom_tools

        self.crewai_tool_descriptions = get_crewai_tools_descriptions()

        self.yaml_config_paths: list[str] = []
        self.agent_tool_mapping: dict[str, list[CrewAITool]] = {}

    def _find_agent_return(self, node: ast.AST) -> Optional[ast.Call]:
        """
        Recursively search for and return the Agent constructor node from a return statement.

        Args:
            node: An AST node to search within

        Returns:
            The Agent constructor node if found, otherwise None
        """
        return find_return_of_function_call(node, self.CREWAI_AGENT_CLASS)

    def _is_agent_constructor(self, node) -> bool:
        """Check if a node is a call to Agent()"""
        return is_function_call(node, self.CREWAI_AGENT_CLASS)

    def _extract_agent_tools(self, agent_node: ast.Call) -> list[CrewAITool]:
        """
        Extract the list of tools from an Agent constructor node.

        Args:
            agent_node: The Agent constructor AST node

        Returns:
            A list of tool names used by this agent
        """
        tools = []

        # Look for the tools parameter in the Agent constructor
        for keyword in agent_node.keywords:
            if keyword.arg == "tools":
                # Found the tools parameter
                tool_list = keyword.value

                if not isinstance(tool_list, ast.List):
                    print(f"Agent tools parameter is not a list: {tool_list}")
                    break

                # Handle tools=[tool1, tool2, ...] format
                for tool_node in tool_list.elts:
                    tool = self._extract_tool(tool_node)
                    if tool:
                        tools.append(tool)

        return tools

    def _extract_tool(self, node: ast.AST) -> Optional[CrewAITool]:
        """
        Extract the tool from a node in the tools list.

        Args:
            node: An AST node representing a tool

        Returns:
            The CrewAITool object if tool is identified identified, otherwise None
        """
        # Handle direct references like MyTools.some_tool
        if isinstance(node, ast.Attribute):
            if node.attr in self.custom_tools:
                return self.custom_tools[node.attr]
            elif node.attr in self.predefined_tools:
                return self.predefined_tools[node.attr]
        # Handle simple names like some_tool
        elif isinstance(node, ast.Name):
            if node.id in self.custom_tools:
                return self.custom_tools[node.id]
            elif node.id in self.predefined_tools:
                return self.predefined_tools[node.id]
        # Handle constructor calls of predefined tools like FileReadTool() or crewai_tools.FileReadTool()
        elif isinstance(node, ast.Call):
            if isinstance(node.func, ast.Name):
                if node.func.id in self.known_tool_aliases:
                    description = self.crewai_tool_descriptions.get(node.func.id, "")
                    return CrewAITool(
                        name=node.func.id, custom=False, description=description
                    )
            elif isinstance(node.func, ast.Attribute):
                if node.func.attr in self.known_tool_aliases:
                    description = self.crewai_tool_descriptions.get(node.func.attr, "")
                    return CrewAITool(
                        name=node.func.attr, custom=False, description=description
                    )

        return None

    def _handle_string_node(self, s: str) -> None:
        if not s.endswith(".yaml") and not s.endswith(".yml"):
            return

        self.yaml_config_paths.append(s)

    def visit_FunctionDef(self, node):
        """Track functions that return an Agent instance."""

        agent_node = self._find_agent_return(node.body)
        if not agent_node:
            self.generic_visit(node)
            return

        agent_tools = self._extract_agent_tools(agent_node)
        self.agent_tool_mapping[node.name] = agent_tools

        self.generic_visit(node)

    def visit_Assign(self, node):
        """Track assignments that return an Agent instance."""
        if not isinstance(node.value, ast.Call):
            self.generic_visit(node)
            return

        if not self._is_agent_constructor(node.value):
            self.generic_visit(node)
            return

        # Handles cases like agent = Agent(...)
        for target in node.targets:
            if isinstance(target, ast.Name):
                agent_tools = self._extract_agent_tools(node.value)
                self.agent_tool_mapping[target.id] = agent_tools

        self.generic_visit(node)

    def visit_Constant(self, node):  # To visit strings in Python >= 3.8
        """Tracks strings that represent path to yaml configuration files."""
        if isinstance(node.value, str):
            self._handle_string_node(node.value)

    def visit_Str(self, node):  # To visit strings in Python < 3.8
        """Tracks strings that represent path to yaml configuration files."""
        self._handle_string_node(node.s)


def collect_agents(
    root_dir: str,
    known_tool_aliases: set[str],
    predefined_tools: dict[str, CrewAITool],
    custom_tools: dict[str, CrewAITool],
) -> dict[str, list[CrewAITool]]:
    """Parses all Python modules in the given directory and collects agents together with their tools.

    Args:
        root_dir (str): Path to the codebase directory
        known_tool_aliases (set[str]): Set of predefined tool aliases parsed from import statements
        predefined_tools (dict[str, CrewAITool]): Dictionary mapping variable name to predefined tool
        custom_tools (dict[str, CrewAITool]): Dictionary mapping variable name to custom tool

    Returns:
        dict[str, list[CrewAITool]]: A dictionary mapping agent names to their tools
    """
    agent_tool_mapping: dict[str, list[CrewAITool]] = {}

    yaml_file_to_agent_tool_mapping = collect_agent_tools_from_config(root_dir)

    for file in walk_python_files(root_dir):
        with open(file, "r") as f:
            try:
                tree = ast.parse(f.read())
            except Exception as e:
                print(f"Cannot parse Python module: {file}. Error: {e}")
                continue
            agents_visitor = AgentsVisitor(
                known_tool_aliases=known_tool_aliases,
                predefined_tools=predefined_tools,
                custom_tools=custom_tools,
            )
            agents_visitor.visit(tree)
            agent_tool_mapping |= agents_visitor.agent_tool_mapping

            # Add agent-tool mapping from YAML config files
            for found_config_path in agents_visitor.yaml_config_paths:
                for (
                    config_path,
                    yaml_agent_tool_mapping,
                ) in yaml_file_to_agent_tool_mapping.items():
                    if found_config_path not in config_path:
                        continue
                    for yaml_agent, yaml_tools in yaml_agent_tool_mapping.items():
                        if (
                            yaml_tools
                            and yaml_agent in agent_tool_mapping
                            and not agent_tool_mapping[yaml_agent]
                        ):
                            agent_tool_mapping[yaml_agent] = yaml_tools

    return agent_tool_mapping
